use std::collections::HashMap;
use std::path::Path;
use std::sync::mpsc::Receiver;
use std::time::{Duration, Instant};

use ferretro_components::base::ControlFlow;
use ferretro_components::prelude::*;

use termion::event::Key;
use termion::input::TermRead;

pub struct TermiosInputComponent {
    //reader: AsyncReader,
    receiver: Receiver<(Instant, Key)>,
    want_quit: bool,
    preferred_pad: Option<u32>,
    button_map: HashMap<Key, InputDeviceId>,
    axis_maps: [HashMap<Key, (i16, i16)>; 2],
    button_state: HashMap<InputDeviceId, (Instant, i16)>,
    axis_state: [(Instant, [i16; 2]); 2],
}

impl Default for TermiosInputComponent {
    fn default() -> Self {
        let (sender, receiver) = std::sync::mpsc::channel();
        std::thread::spawn(move || {
            for k in std::io::stdin().keys().map(|kr| kr.unwrap()) {
                sender.send((Instant::now(), k)).unwrap();
                if k == Key::Esc {
                    break
                }
            }
        });
        TermiosInputComponent {
            //reader: termion::async_stdin(),
            receiver,
            want_quit: false,
            preferred_pad: None,
            button_map: [
                (Key::Up, InputDeviceId::Joypad(JoypadButton::Up)),
                (Key::Down, InputDeviceId::Joypad(JoypadButton::Down)),
                (Key::Left, InputDeviceId::Joypad(JoypadButton::Left)),
                (Key::Right, InputDeviceId::Joypad(JoypadButton::Right)),
                (Key::Char('x'), InputDeviceId::Joypad(JoypadButton::A)),
                (Key::Char('z'), InputDeviceId::Joypad(JoypadButton::B)),
                (Key::Char('s'), InputDeviceId::Joypad(JoypadButton::X)),
                (Key::Char('a'), InputDeviceId::Joypad(JoypadButton::Y)),
                (Key::Ctrl('x'), InputDeviceId::Joypad(JoypadButton::A)),
                (Key::Ctrl('z'), InputDeviceId::Joypad(JoypadButton::B)),
                (Key::Ctrl('s'), InputDeviceId::Joypad(JoypadButton::X)),
                (Key::Ctrl('a'), InputDeviceId::Joypad(JoypadButton::Y)),
                (Key::Alt('x'), InputDeviceId::Joypad(JoypadButton::A)),
                (Key::Alt('z'), InputDeviceId::Joypad(JoypadButton::B)),
                (Key::Alt('s'), InputDeviceId::Joypad(JoypadButton::X)),
                (Key::Alt('a'), InputDeviceId::Joypad(JoypadButton::Y)),
                (Key::Char('q'), InputDeviceId::Joypad(JoypadButton::L)),
                (Key::Char('w'), InputDeviceId::Joypad(JoypadButton::R)),
                (Key::Ctrl('q'), InputDeviceId::Joypad(JoypadButton::L2)),
                (Key::Ctrl('w'), InputDeviceId::Joypad(JoypadButton::R2)),
                (Key::Alt('q'), InputDeviceId::Joypad(JoypadButton::L3)),
                (Key::Alt('w'), InputDeviceId::Joypad(JoypadButton::R3)),
                (Key::Char('\n'), InputDeviceId::Joypad(JoypadButton::Start)),
                (Key::Backspace, InputDeviceId::Joypad(JoypadButton::Select)),
                (Key::Char('\t'), InputDeviceId::Joypad(JoypadButton::Select)),
            ].into_iter().collect(),
            axis_maps: [
                [
                    (Key::Char('1'), (i16::MIN, i16::MAX)),
                    (Key::Char('2'), (0, i16::MAX)),
                    (Key::Char('3'), (i16::MAX, i16::MAX)),
                    (Key::Char('4'), (i16::MIN, 0)),
                    (Key::Char('5'), (0, 0)),
                    (Key::Char('6'), (i16::MAX, 0)),
                    (Key::Char('7'), (i16::MIN + 1, i16::MIN + 1)), // ???
                    (Key::Char('8'), (0, i16::MIN)),
                    (Key::Char('9'), (i16::MAX, i16::MIN)),
                ].into_iter().collect(),
                [
                    (Key::Char('r'), (0, i16::MIN)),
                    (Key::Char('d'), (0, i16::MAX)),
                    (Key::Char('e'), (i16::MIN, 0)),
                    (Key::Char('f'), (i16::MAX, 0)),
                ].into_iter().collect(),
            ],
            button_state: Default::default(),
            axis_state: [(Instant::now(), [0, 0]); 2],
        }
    }
}

impl RetroCallbacks for TermiosInputComponent {
    fn input_poll(&mut self) {
        while let Ok((now, k)) = self.receiver.try_recv() {
            if k == Key::Esc {
                self.want_quit = true;
            }

            if let Some(mapping) = self.button_map.get(&k) {
                self.button_state.insert(mapping.to_owned(), (now, 1));
            }

            for (axis_map, axis_state) in self.axis_maps.iter().zip(self.axis_state.iter_mut()) {
                let (mut sum_x, mut sum_y) = (0, 0);
                let mut count = 0;
                if let Some((x, y)) = axis_map.get(&k) {
                    sum_x += *x as i32;
                    sum_y += *y as i32;
                    count += 1;
                }
                if count != 0 {
                    let average_x = (sum_x / count) as i16;
                    let average_y = (sum_y / count) as i16;
                    *axis_state = (now, [average_x, average_y]);
                }
            }
        }
    }

    fn input_state(&mut self, port: u32, device: InputDeviceId, index: InputIndex) -> i16 {
        if port != 0 {
            return 0;
        }
        if let Some((inst, val)) = self.button_state.get(&device) {
            // TODO: consult kbdrate.c (and X11?) for durations (can't detect key-up/held, only repeat)
            if Instant::now().duration_since(*inst) < Duration::from_millis(300) {
                return *val;
            }
        }
        match device {
            InputDeviceId::Analog(axis_id) => {
                let (inst, axes) = self.axis_state[index as u32 as usize];
                let since = Instant::now().duration_since(inst);
                let ratio = if since < Duration::from_millis(100) {
                    1.0
                } else if since < Duration::from_millis(300) {
                    (0.3 - since.as_secs_f32()) * 5.0
                } else {
                    0.0
                };
                (axes[axis_id as u32 as usize] as f32 * ratio) as i16
            }
            // TODO: mouse?
            _ => 0,
        }
    }

    fn get_variable(&mut self, key: &str) -> Option<String> {
        match key {
            "beetle_saturn_analog_stick_deadzone" => Some("15%".to_string()),
            "parallel-n64-astick-deadzone" => Some("15%".to_string()),
            _ => None,
        }
    }

    fn get_input_device_capabilities(&mut self) -> Option<u64> {
        let bits = (1 << (DeviceType::Joypad as u32)) | (1 << (DeviceType::Analog as u32));
        Some(bits as u64)
    }

    fn set_controller_info(&mut self, controller_info: &[ControllerDescription2]) -> Option<bool> {
        for ci in controller_info {
            // so we can have analog support in beetle/mednafen saturn
            if ci.name.as_str() == "3D Control Pad" {
                self.preferred_pad = Some(ci.device_id());
                break;
            }
        }
        Some(true)
    }
}

impl RetroComponent for TermiosInputComponent {
    fn post_run(&mut self, _: &mut LibretroWrapper) -> ControlFlow {
        if self.want_quit {
            ControlFlow::Break
        } else {
            ControlFlow::Continue
        }
    }

    fn post_load_game(&mut self, retro: &mut LibretroWrapper, _rom: &Path) -> ferretro_components::base::Result<()> {
        if let Some(device) = self.preferred_pad {
            retro.set_controller_port_device(0, device);
        }
        Ok(())
    }
}
